import numpy as np
from valor_lite.detection import (
    DataLoader,
    Detection,
    Evaluator,
    compute_detailed_pr_curve,
)


def test_detailed_pr_curve_no_data():
    evaluator = Evaluator()
    curves = evaluator.compute_detailed_pr_curve()
    assert isinstance(curves, list)
    assert len(curves) == 0


def test_compute_detailed_pr_curve():
    sorted_pairs = np.array(
        [
            # dt,  gt,  pd,  iou,  gl,  pl, score,
            [0.0, 0.0, 1.0, 0.98, 0.0, 0.0, 0.95],
            [1.0, 1.0, 2.0, 0.55, 1.0, 0.0, 0.95],
            [2.0, -1.0, 3.0, 0.67, -1.0, 0.0, 0.65],
            [3.0, 4.0, 4.0, 1.0, 0.0, 0.0, 0.1],
            [4.0, 5.0, -1.0, 0.5, 0.0, -1.0, -1.0],
        ]
    )
    label_counts = np.array([[3, 4], [1, 0]])
    iou_thresholds = np.array([0.5])
    score_thresholds = np.array([score / 100.0 for score in range(1, 101)])

    results = compute_detailed_pr_curve(
        data=sorted_pairs,
        label_counts=label_counts,
        iou_thresholds=iou_thresholds,
        score_thresholds=score_thresholds,
        n_samples=0,
    )

    assert len(results) == 1
    assert results.shape == (1, 100, 2, 5)  # iou, score, label, metrics

    """
    @ iou=0.5, score<0.1
    2x tp
    1x fp misclassification
    1x fp hallucination
    0x fn misclassification
    1x fn missing prediction
    """
    assert np.isclose(results[0, :10, 0, :], np.array([2, 1, 1, 0, 1])).all()

    """
    @ iou=0.5, score=0.5
    1x tp
    1x fp misclassification
    1x fp hallucination
    1x fn misclassification
    1x fn missing prediction
    """
    assert np.isclose(results[0, 10:95, 0, :], np.array([1, 1, 1, 1, 1])).all()

    """
    @ iou=0.5, score>=0.95
    0x tp
    0x fp misclassification
    2x fp hallucination
    2x fn misclassification
    1x fn missing prediction
    """
    assert np.isclose(results[0, 95:, 0, :], np.array([0, 0, 2, 2, 1])).all()

    # compute with examples

    """

    output

    label_idx
    tp
    ... examples
    fp_misclassification
    ... examples
    fp_hallucination
    ... examples
    fn_misclassification
    ... examples
    fn_missing_prediction
    ... examples
    """

    n_samples = 2

    results = compute_detailed_pr_curve(
        data=sorted_pairs,
        label_counts=label_counts,
        iou_thresholds=iou_thresholds,
        score_thresholds=score_thresholds,
        n_samples=n_samples,
    )

    assert len(results) == 1
    assert results.shape == (1, 100, 2, 15)  # iou, score, label, metrics

    tp_idx = 0
    fp_misclf_idx = tp_idx + n_samples + 1
    fp_halluc_idx = fp_misclf_idx + n_samples + 1
    fn_misclf_idx = fp_halluc_idx + n_samples + 1
    fn_misprd_idx = fn_misclf_idx + n_samples + 1

    metric_indices = np.zeros((15,), dtype=bool)
    for index in [
        tp_idx,
        fp_misclf_idx,
        fp_halluc_idx,
        fn_misclf_idx,
        fn_misprd_idx,
    ]:
        metric_indices[index] = True

    """
    @ iou=0.5, score<0.1
    2x tp
    1x fp misclassification
    1x fp hallucination
    0x fn misclassification
    1x fn missing prediction
    """
    assert np.isclose(
        results[0, :10, 0, metric_indices],
        np.array([2, 1, 1, 0, 1])[:, np.newaxis],
    ).all()  # metrics
    assert np.isclose(
        results[0, :10, 0, tp_idx + 1 : fp_misclf_idx], np.array([0.0, 3.0])
    ).all()  # tp
    assert np.isclose(
        results[0, :10, 0, fp_misclf_idx + 1 : fp_halluc_idx],
        np.array([1.0, -1.0]),
    ).all()  # fp misclf
    assert np.isclose(
        results[0, :10, 0, fp_halluc_idx + 1 : fn_misclf_idx],
        np.array([2.0, -1.0]),
    ).all()  # fp halluc
    assert np.isclose(
        results[0, :10, 0, fn_misclf_idx + 1 : fn_misprd_idx],
        np.array([-1.0, -1.0]),
    ).all()  # fn misclf
    assert np.isclose(
        results[0, :10, 0, fn_misprd_idx + 1 :], np.array([4.0, -1.0])
    ).all()  # fn misprd

    """
    @ iou=0.5, score=0.5
    1x tp
    1x fp misclassification
    1x fp hallucination
    1x fn misclassification
    1x fn missing prediction
    """
    assert np.isclose(
        results[0, 10:95, 0, metric_indices],
        np.array([1, 1, 1, 1, 1])[:, np.newaxis],
    ).all()
    assert np.isclose(
        results[0, 10:95, 0, tp_idx + 1 : fp_misclf_idx], np.array([0.0, -1.0])
    ).all()  # tp
    assert np.isclose(
        results[0, 10:95, 0, fp_misclf_idx + 1 : fp_halluc_idx],
        np.array([1.0, -1.0]),
    ).all()  # fp misclf
    assert np.isclose(
        results[0, 10:95, 0, fp_halluc_idx + 1 : fn_misclf_idx],
        np.array([2.0, -1.0]),
    ).all()  # fp halluc
    assert np.isclose(
        results[0, 10:95, 0, fn_misclf_idx + 1 : fn_misprd_idx],
        np.array([3.0, -1.0]),
    ).all()  # fn misclf
    assert np.isclose(
        results[0, 10:95, 0, fn_misprd_idx + 1 :], np.array([4.0, -1.0])
    ).all()  # fn misprd

    """
    @ iou=0.5, score>=0.95
    0x tp
    0x fp misclassification
    2x fp hallucination
    2x fn misclassification
    1x fn missing prediction
    """
    assert np.isclose(
        results[0, 95:, 0, metric_indices],
        np.array([0, 0, 2, 2, 1])[:, np.newaxis],
    ).all()
    assert np.isclose(
        results[0, 95:, 0, tp_idx + 1 : fp_misclf_idx], np.array([-1.0, -1.0])
    ).all()  # tp
    assert np.isclose(
        results[0, 95:, 0, fp_misclf_idx + 1 : fp_halluc_idx],
        np.array([-1.0, -1.0]),
    ).all()  # fp misclf
    assert np.isclose(
        results[0, 95:, 0, fp_halluc_idx + 1 : fn_misclf_idx],
        np.array([1.0, 2.0]),
    ).all()  # fp halluc
    assert np.isclose(
        results[0, 95:, 0, fn_misclf_idx + 1 : fn_misprd_idx],
        np.array([0.0, 3.0]),
    ).all()  # fn misclf
    assert np.isclose(
        results[0, 95:, 0, fn_misprd_idx + 1 :], np.array([4.0, -1.0])
    ).all()  # fn misprd


def test_detailed_pr_curve_using_torch_metrics_example(
    torchmetrics_detections: list[Detection],
):
    """
    cf with torch metrics/pycocotools results listed here:
    https://github.com/Lightning-AI/metrics/blob/107dbfd5fb158b7ae6d76281df44bd94c836bfce/tests/unittests/detection/test_map.py#L231
    """
    manager = DataLoader()
    manager.add_data(torchmetrics_detections)
    evaluator = manager.finalize()

    assert evaluator.ignored_prediction_labels == [("class", "3")]
    assert evaluator.missing_prediction_labels == []
    assert evaluator.n_datums == 4
    assert evaluator.n_labels == 6
    assert evaluator.n_groundtruths == 20
    assert evaluator.n_predictions == 19

    metrics = evaluator.compute_detailed_pr_curve(
        iou_thresholds=[0.5, 0.75],
        score_thresholds=[0.25, 0.75],
        n_samples=1,
    )

    # test DetailedPrecisionRecallCurve
    actual_metrics = [m.to_dict() for m in metrics]
    expected_metrics = [
        {
            "value": [
                {
                    "score": 0.25,
                    "tp": 1.0,
                    "fp_misclassification": 0.0,
                    "fp_hallucination": 0.0,
                    "fn_misclassification": 1.0,
                    "fn_missing_prediction": 0.0,
                    "tp_examples": ["2"],
                    "fp_misclassification_examples": [],
                    "fp_hallucination_examples": [],
                    "fn_misclassification_examples": ["0"],
                    "fn_missing_prediction_examples": [],
                },
                {
                    "score": 0.75,
                    "tp": 0.0,
                    "fp_misclassification": 0.0,
                    "fp_hallucination": 6.0,
                    "fn_misclassification": 2.0,
                    "fn_missing_prediction": 4.0,
                    "tp_examples": [],
                    "fp_misclassification_examples": [],
                    "fp_hallucination_examples": ["2"],
                    "fn_misclassification_examples": ["0"],
                    "fn_missing_prediction_examples": ["2"],
                },
            ],
            "iou": 0.5,
            "label": {"key": "class", "value": "4"},
            "type": "DetailedPrecisionRecallCurve",
        },
        {
            "value": [
                {
                    "score": 0.25,
                    "tp": 1.0,
                    "fp_misclassification": 0.0,
                    "fp_hallucination": 0.0,
                    "fn_misclassification": 1.0,
                    "fn_missing_prediction": 0.0,
                    "tp_examples": ["2"],
                    "fp_misclassification_examples": [],
                    "fp_hallucination_examples": [],
                    "fn_misclassification_examples": ["0"],
                    "fn_missing_prediction_examples": [],
                },
                {
                    "score": 0.75,
                    "tp": 0.0,
                    "fp_misclassification": 0.0,
                    "fp_hallucination": 6.0,
                    "fn_misclassification": 2.0,
                    "fn_missing_prediction": 4.0,
                    "tp_examples": [],
                    "fp_misclassification_examples": [],
                    "fp_hallucination_examples": ["2"],
                    "fn_misclassification_examples": ["0"],
                    "fn_missing_prediction_examples": ["2"],
                },
            ],
            "iou": 0.75,
            "label": {"key": "class", "value": "4"},
            "type": "DetailedPrecisionRecallCurve",
        },
        {
            "value": [
                {
                    "score": 0.25,
                    "tp": 1.0,
                    "fp_misclassification": 0.0,
                    "fp_hallucination": 1.0,
                    "fn_misclassification": 0.0,
                    "fn_missing_prediction": 1.0,
                    "tp_examples": ["1"],
                    "fp_misclassification_examples": [],
                    "fp_hallucination_examples": ["1"],
                    "fn_misclassification_examples": [],
                    "fn_missing_prediction_examples": ["1"],
                },
                {
                    "score": 0.75,
                    "tp": 0.0,
                    "fp_misclassification": 0.0,
                    "fp_hallucination": 1.0,
                    "fn_misclassification": 1.0,
                    "fn_missing_prediction": 3.0,
                    "tp_examples": [],
                    "fp_misclassification_examples": [],
                    "fp_hallucination_examples": ["1"],
                    "fn_misclassification_examples": ["1"],
                    "fn_missing_prediction_examples": ["1"],
                },
            ],
            "iou": 0.5,
            "label": {"key": "class", "value": "2"},
            "type": "DetailedPrecisionRecallCurve",
        },
        {
            "value": [
                {
                    "score": 0.25,
                    "tp": 1.0,
                    "fp_misclassification": 0.0,
                    "fp_hallucination": 1.0,
                    "fn_misclassification": 0.0,
                    "fn_missing_prediction": 1.0,
                    "tp_examples": ["1"],
                    "fp_misclassification_examples": [],
                    "fp_hallucination_examples": ["1"],
                    "fn_misclassification_examples": [],
                    "fn_missing_prediction_examples": ["1"],
                },
                {
                    "score": 0.75,
                    "tp": 0.0,
                    "fp_misclassification": 0.0,
                    "fp_hallucination": 1.0,
                    "fn_misclassification": 1.0,
                    "fn_missing_prediction": 3.0,
                    "tp_examples": [],
                    "fp_misclassification_examples": [],
                    "fp_hallucination_examples": ["1"],
                    "fn_misclassification_examples": ["1"],
                    "fn_missing_prediction_examples": ["1"],
                },
            ],
            "iou": 0.75,
            "label": {"key": "class", "value": "2"},
            "type": "DetailedPrecisionRecallCurve",
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


# @pytest.fixture
# def test_detailed_precision_recall_curve(
#     evaluate_detection_detailed_pr_curve_groundtruths: list,
#     evaluate_detection_detailed_pr_curve_predictions: list,
#     detailed_precision_recall_curve_outputs: tuple,
# ):

#     expected_outputs, _ = detailed_precision_recall_curve_outputs

#     Dataloader = Dataloader(
#         metrics_to_return=[enums.MetricType.DetailedDetailedPrecisionRecallCurve],
#     )

#     Dataloader.add_data(
#         groundtruths=evaluate_detection_detailed_pr_curve_groundtruths,
#         predictions=evaluate_detection_detailed_pr_curve_predictions,
#     )

#     # check that ious have been precomputed
#     assert "iou_" in Dataloader.joint_df.columns
#     assert all(
#         [
#             col not in ["raster", "bounding_box"]
#             for col in Dataloader.joint_df.columns
#         ]
#     )

#     eval_job = Dataloader.evaluate()
#     for key, expected_value in expected_outputs.items():
#         result = eval_job.metrics[0]["value"]
#         for k in key:
#             result = result[k]
#         assert result == expected_value

#     # repeat tests using a lower IOU threshold
#     Dataloader = Dataloader(
#         metrics_to_return=[enums.MetricType.DetailedDetailedPrecisionRecallCurve],
#         pr_curve_iou_threshold=0.45,
#     )

#     Dataloader.add_data(
#         groundtruths=evaluate_detection_detailed_pr_curve_groundtruths,
#         predictions=evaluate_detection_detailed_pr_curve_predictions,
#     )

#     eval_job_low_iou_threshold = Dataloader.evaluate()

#     for key, expected_value in expected_outputs.items():
#         result = eval_job_low_iou_threshold.metrics[0]["value"]
#         for k in key:
#             result = result[k]
#         assert result == expected_value


# def test_evaluate_detection_model_with_no_predictions(
#     evaluate_detection_groundtruths: list,
#     evaluate_detection_model_with_no_predictions_output: list,
# ):
#     """
#     Test detection evaluations when the model outputs nothing.

#     gt_dets1
#         datum 1
#             - Label (k1, v1) with Annotation area = 1500
#             - Label (k2, v2) with Annotation area = 57,510
#         datum2
#             - Label (k1, v1) with Annotation area = 1100
#     """
#     predictions = []
#     for gt in evaluate_detection_groundtruths:
#         predictions.append(
#             schemas.Prediction(
#                 datum=gt.datum,
#                 annotations=[],
#             )
#         )

#     Dataloader = Dataloader()

#     # can't pass empty lists, but can pass predictions without annotations
#     with pytest.raises(ValueError) as e:
#         Dataloader.add_data(
#             groundtruths=evaluate_detection_groundtruths,
#             predictions=[],
#         )
#     assert (
#         "it's neither a dataframe nor a list of Valor Prediction objects"
#         in str(e)
#     )

#     Dataloader.add_data(
#         groundtruths=evaluate_detection_groundtruths,
#         predictions=predictions,
#     )

#     # check that ious have been precomputed
#     assert "iou_" in Dataloader.joint_df.columns
#     assert all(
#         [
#             col not in ["raster", "bounding_box"]
#             for col in Dataloader.joint_df.columns
#         ]
#     )

#     eval_job = Dataloader.evaluate()

#     computed_metrics = eval_job.metrics

#     assert all([metric["value"] == 0 for metric in computed_metrics])

#     for m in evaluate_detection_model_with_no_predictions_output:
#         assert m in computed_metrics

#     for m in computed_metrics:
#         assert m in evaluate_detection_model_with_no_predictions_output


# def test_evaluate_detection_functional_test(
#     evaluate_detection_functional_test_groundtruths: list,
#     evaluate_detection_functional_test_predictions: list,
#     evaluate_detection_functional_test_outputs: tuple,
# ):

#     (
#         expected_metrics,
#         pr_expected_answers,
#         detailed_pr_expected_answers,
#         higher_iou_threshold_pr_expected_answers,
#         higher_iou_threshold_detailed_pr_expected_answers,
#     ) = evaluate_detection_functional_test_outputs

#     Dataloader = Dataloader()
#     Dataloader.add_data(
#         groundtruths=evaluate_detection_functional_test_groundtruths,
#         predictions=evaluate_detection_functional_test_predictions,
#     )
#     Dataloader.finalize()

#     ap_metrics = translate_ap_metrics(
#         Dataloader.compute_ap(
#             iou_thresholds=[0.5, 0.75]
#         )
#     )

#     pr_curves = translate_pr_curves(
#         Dataloader.compute_pr_curve(
#             iou_thresholds=[0.5],
#             n_samples=1,
#         )
#     )

#         metrics_to_return=[
#             enums.MetricType.AP,
#             enums.MetricType.AR,
#             enums.MetricType.mAP,
#             enums.MetricType.APAveragedOverIOUs,
#             enums.MetricType.mAR,
#             enums.MetricType.mAPAveragedOverIOUs,
#             enums.MetricType.DetailedPrecisionRecallCurve,
#             enums.MetricType.DetailedDetailedPrecisionRecallCurve,
#         ],
#         pr_curve_iou_threshold=0.5,
#         pr_curve_max_examples=1,
#     )


#     metrics = [
#         m
#         for m in eval_job.metrics
#         if m["type"]
#         not in ["DetailedPrecisionRecallCurve", "DetailedDetailedPrecisionRecallCurve"]
#     ]

#     # round all metrics to the third decimal place
#     for i, m in enumerate(metrics):
#         metrics[i]["value"] = round(m["value"], 3)

#     pr_metrics = [
#         m for m in eval_job.metrics if m["type"] == "DetailedPrecisionRecallCurve"
#     ]
#     detailed_pr_metrics = [
#         m
#         for m in eval_job.metrics
#         if m["type"] == "DetailedDetailedPrecisionRecallCurve"
#     ]

#     for m in metrics:
#         assert m in expected_metrics
#     for m in metrics:
#         assert m in eval_job.metrics

#     for (
#         _,
#         value,
#         threshold,
#         metric,
#     ), expected_value in pr_expected_answers.items():
#         assert (
#             pr_metrics[0]["value"][value][threshold][metric] == expected_value
#         )

#     for (
#         value,
#         threshold,
#         metric,
#     ), expected_output in detailed_pr_expected_answers.items():
#         model_output = detailed_pr_metrics[0]["value"][value][threshold][
#             metric
#         ]
#         assert isinstance(model_output, dict)
#         assert model_output["total"] == expected_output["total"]
#         assert all(
#             [
#                 model_output["observations"][key]["count"]  # type: ignore - we know this element is a dict
#                 == expected_output[key]
#                 for key in [
#                     key
#                     for key in expected_output.keys()
#                     if key not in ["total"]
#                 ]
#             ]
#         )

#     # spot check number of examples
#     assert (
#         len(
#             detailed_pr_metrics[0]["value"]["0"][0.95]["fn"]["observations"]["no_predictions"][  # type: ignore - we know this element is a dict
#                 "examples"
#             ]
#         )
#         == 1
#     )
#     assert (
#         len(
#             detailed_pr_metrics[0]["value"]["49"][0.05]["tp"]["observations"]["all"][  # type: ignore - we know this element is a dict
#                 "examples"
#             ]
#         )
#         == 1
#     )

#     # raise the iou threshold
#     Dataloader = Dataloader(
#         metrics_to_return=[
#             enums.MetricType.DetailedPrecisionRecallCurve,
#             enums.MetricType.DetailedDetailedPrecisionRecallCurve,
#         ],
#         pr_curve_iou_threshold=0.9,
#         pr_curve_max_examples=1,
#     )

#     Dataloader.add_data(
#         groundtruths=evaluate_detection_functional_test_groundtruths,
#         predictions=evaluate_detection_functional_test_predictions,
#     )

#     # check that ious have been precomputed
#     assert "iou_" in Dataloader.joint_df.columns
#     assert all(
#         [
#             col not in ["raster", "bounding_box"]
#             for col in Dataloader.joint_df.columns
#         ]
#     )

#     eval_job_higher_threshold = Dataloader.evaluate()

#     pr_metrics = [
#         m
#         for m in eval_job_higher_threshold.metrics
#         if m["type"] == "DetailedPrecisionRecallCurve"
#     ]
#     detailed_pr_metrics = [
#         m
#         for m in eval_job_higher_threshold.metrics
#         if m["type"] == "DetailedDetailedPrecisionRecallCurve"
#     ]

#     for (
#         key,
#         value,
#         threshold,
#         metric,
#     ), expected_count in higher_iou_threshold_pr_expected_answers.items():
#         actual_count = pr_metrics[0]["value"][value][threshold][metric]
#         assert actual_count == expected_count

#     for (
#         value,
#         threshold,
#         metric,
#     ), expected_output in (
#         higher_iou_threshold_detailed_pr_expected_answers.items()
#     ):
#         model_output = detailed_pr_metrics[0]["value"][value][threshold][
#             metric
#         ]
#         assert isinstance(model_output, dict)
#         assert model_output["total"] == expected_output["total"]
#         assert all(
#             [
#                 model_output["observations"][key]["count"]  # type: ignore - we know this element is a dict
#                 == expected_output[key]
#                 for key in [
#                     key
#                     for key in expected_output.keys()
#                     if key not in ["total"]
#                 ]
#             ]
#         )

#     assert (
#         len(
#             detailed_pr_metrics[0]["value"]["0"][0.95]["fn"]["observations"]["no_predictions"][  # type: ignore - we know this element is a dict
#                 "examples"
#             ]
#         )
#         == 1
#     )
#     assert (
#         len(
#             detailed_pr_metrics[0]["value"]["49"][0.05]["tp"]["observations"]["all"][  # type: ignore - we know this element is a dict
#                 "examples"
#             ]
#         )
#         == 1
#     )

#     # repeat the above, but with a higher pr_max_curves_example
#     Dataloader = Dataloader(
#         metrics_to_return=[
#             enums.MetricType.DetailedPrecisionRecallCurve,
#             enums.MetricType.DetailedDetailedPrecisionRecallCurve,
#         ],
#         pr_curve_iou_threshold=0.9,
#         pr_curve_max_examples=3,
#     )

#     Dataloader.add_data(
#         groundtruths=evaluate_detection_functional_test_groundtruths,
#         predictions=evaluate_detection_functional_test_predictions,
#     )

#     # check that ious have been precomputed
#     assert "iou_" in Dataloader.joint_df.columns
#     assert all(
#         [
#             col not in ["raster", "bounding_box"]
#             for col in Dataloader.joint_df.columns
#         ]
#     )

#     eval_job_higher_threshold = Dataloader.evaluate()

#     pr_metrics = [
#         m
#         for m in eval_job_higher_threshold.metrics
#         if m["type"] == "DetailedPrecisionRecallCurve"
#     ]
#     detailed_pr_metrics = [
#         m
#         for m in eval_job_higher_threshold.metrics
#         if m["type"] == "DetailedDetailedPrecisionRecallCurve"
#     ]

#     for (
#         key,
#         value,
#         threshold,
#         metric,
#     ), expected_count in higher_iou_threshold_pr_expected_answers.items():
#         actual_count = pr_metrics[0]["value"][value][threshold][metric]
#         assert actual_count == expected_count

#     for (
#         value,
#         threshold,
#         metric,
#     ), expected_output in (
#         higher_iou_threshold_detailed_pr_expected_answers.items()
#     ):
#         model_output = detailed_pr_metrics[0]["value"][value][threshold][
#             metric
#         ]
#         assert isinstance(model_output, dict)
#         assert model_output["total"] == expected_output["total"]
#         assert all(
#             [
#                 model_output["observations"][key]["count"]  # type: ignore - we know this element is a dict
#                 == expected_output[key]
#                 for key in [
#                     key
#                     for key in expected_output.keys()
#                     if key not in ["total"]
#                 ]
#             ]
#         )

#     assert (
#         len(
#             detailed_pr_metrics[0]["value"]["0"][0.95]["fn"]["observations"]["no_predictions"][  # type: ignore - we know this element is a dict
#                 "examples"
#             ]
#         )
#         == 3
#     )
#     assert (
#         len(
#             detailed_pr_metrics[0]["value"]["49"][0.05]["tp"]["observations"]["all"][  # type: ignore - we know this element is a dict
#                 "examples"
#             ]
#         )
#         == 2
#     )

#     # test behavior if pr_curve_max_examples == 0
#     Dataloader = Dataloader(
#         metrics_to_return=[
#             enums.MetricType.DetailedPrecisionRecallCurve,
#             enums.MetricType.DetailedDetailedPrecisionRecallCurve,
#         ],
#         pr_curve_iou_threshold=0.9,
#         pr_curve_max_examples=0,
#     )

#     Dataloader.add_data(
#         groundtruths=evaluate_detection_functional_test_groundtruths,
#         predictions=evaluate_detection_functional_test_predictions,
#     )

#     # check that ious have been precomputed
#     assert "iou_" in Dataloader.joint_df.columns
#     assert all(
#         [
#             col not in ["raster", "bounding_box"]
#             for col in Dataloader.joint_df.columns
#         ]
#     )

#     eval_job_higher_threshold = Dataloader.evaluate()

#     pr_metrics = [
#         m
#         for m in eval_job_higher_threshold.metrics
#         if m["type"] == "DetailedPrecisionRecallCurve"
#     ]
#     detailed_pr_metrics = [
#         m
#         for m in eval_job_higher_threshold.metrics
#         if m["type"] == "DetailedDetailedPrecisionRecallCurve"
#     ]

#     for (
#         key,
#         value,
#         threshold,
#         metric,
#     ), expected_count in higher_iou_threshold_pr_expected_answers.items():
#         actual_count = pr_metrics[0]["value"][value][threshold][metric]
#         assert actual_count == expected_count

#     for (
#         value,
#         threshold,
#         metric,
#     ), expected_output in (
#         higher_iou_threshold_detailed_pr_expected_answers.items()
#     ):
#         model_output = detailed_pr_metrics[0]["value"][value][threshold][
#             metric
#         ]
#         assert isinstance(model_output, dict)
#         assert model_output["total"] == expected_output["total"]
#         assert all(
#             [
#                 model_output["observations"][key]["count"]  # type: ignore - we know this element is a dict
#                 == expected_output[key]
#                 for key in [
#                     key
#                     for key in expected_output.keys()
#                     if key not in ["total"]
#                 ]
#             ]
#         )

#     # spot check number of examples
#     assert (
#         len(
#             detailed_pr_metrics[0]["value"]["0"][0.95]["fn"]["observations"]["no_predictions"][  # type: ignore - we know this element is a dict
#                 "examples"
#             ]
#         )
#         == 0
#     )
#     assert (
#         len(
#             detailed_pr_metrics[0]["value"]["49"][0.05]["tp"]["observations"]["all"][  # type: ignore - we know this element is a dict
#                 "examples"
#             ]
#         )
#         == 0
#     )



import pytest
from valor_lite.detection import BoundingBox, Detection


@pytest.fixture
def rect1() -> tuple[float, float, float, float]:
    """Box with area = 1500."""
    return (10.0, 60.0, 10.0, 40.0)


@pytest.fixture
def rect2() -> tuple[float, float, float, float]:
    """Box with area = 1100."""
    return (15.0, 70.0, 0.0, 20.0)


@pytest.fixture
def rect3() -> tuple[float, float, float, float]:
    """Box with area = 57,510."""
    return (87.0, 158.0, 10.0, 820.0)


@pytest.fixture
def rect4() -> tuple[float, float, float, float]:
    """Box with area = 90."""
    return (1.0, 10.0, 10.0, 20.0)


@pytest.fixture
def rect5() -> tuple[float, float, float, float]:
    """Box with partial overlap to rect3."""
    return (87, 158, 10, 400)


@pytest.fixture
def basic_detections(
    rect1: tuple[float, float, float, float],
    rect2: tuple[float, float, float, float],
    rect3: tuple[float, float, float, float],
) -> list[Detection]:
    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=rect1[0],
                    xmax=rect1[1],
                    ymin=rect1[2],
                    ymax=rect1[3],
                    labels=[("k1", "v1")],
                ),
                BoundingBox(
                    xmin=rect3[0],
                    xmax=rect3[1],
                    ymin=rect3[2],
                    ymax=rect3[3],
                    labels=[("k2", "v2")],
                ),
            ],
            predictions=[
                BoundingBox(
                    xmin=rect1[0],
                    xmax=rect1[1],
                    ymin=rect1[2],
                    ymax=rect1[3],
                    labels=[("k1", "v1")],
                    scores=[0.3],
                ),
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[
                BoundingBox(
                    xmin=rect2[0],
                    xmax=rect2[1],
                    ymin=rect2[2],
                    ymax=rect2[3],
                    labels=[("k1", "v1")],
                ),
            ],
            predictions=[
                BoundingBox(
                    xmin=rect2[0],
                    xmax=rect2[1],
                    ymin=rect2[2],
                    ymax=rect2[3],
                    labels=[("k2", "v2")],
                    scores=[0.98],
                ),
            ],
        ),
    ]


@pytest.fixture
def torchmetrics_detections() -> list[Detection]:
    """Creates a model called "test_model" with some predicted
    detections on the dataset "test_dataset". These predictions are taken
    from a torchmetrics unit test (see test_metrics.py)
    """

    # predictions for four images taken from
    # https://github.com/Lightning-AI/metrics/blob/107dbfd5fb158b7ae6d76281df44bd94c836bfce/tests/unittests/detection/test_map.py#L59

    groundtruths = [
        {"boxes": [[214.1500, 41.2900, 562.4100, 285.0700]], "labels": ["4"]},
        {
            "boxes": [
                [13.00, 22.75, 548.98, 632.42],
                [1.66, 3.32, 270.26, 275.23],
            ],
            "labels": ["2", "2"],
        },
        {
            "boxes": [
                [61.87, 276.25, 358.29, 379.43],
                [2.75, 3.66, 162.15, 316.06],
                [295.55, 93.96, 313.97, 152.79],
                [326.94, 97.05, 340.49, 122.98],
                [356.62, 95.47, 372.33, 147.55],
                [462.08, 105.09, 493.74, 146.99],
                [277.11, 103.84, 292.44, 150.72],
            ],
            "labels": ["4", "1", "0", "0", "0", "0", "0"],
        },
        {
            "boxes": [
                [72.92, 45.96, 91.23, 80.57],
                [50.17, 45.34, 71.28, 79.83],
                [81.28, 47.04, 98.66, 78.50],
                [63.96, 46.17, 84.35, 80.48],
                [75.29, 23.01, 91.85, 50.85],
                [56.39, 21.65, 75.66, 45.54],
                [73.14, 1.10, 98.96, 28.33],
                [62.34, 55.23, 78.14, 79.57],
                [44.17, 45.78, 63.99, 78.48],
                [58.18, 44.80, 66.42, 56.25],
            ],
            "labels": [
                "49",
                "49",
                "49",
                "49",
                "49",
                "49",
                "49",
                "49",
                "49",
                "49",
            ],
        },
    ]
    predictions = [
        {
            "boxes": [[258.15, 41.29, 606.41, 285.07]],
            "scores": [0.236],
            "labels": ["4"],
        },
        {
            "boxes": [
                [61.00, 22.75, 565.00, 632.42],
                [12.66, 3.32, 281.26, 275.23],
            ],
            "scores": [0.318, 0.726],
            "labels": ["3", "2"],
        },
        {
            "boxes": [
                [87.87, 276.25, 384.29, 379.43],
                [0.00, 3.66, 142.15, 316.06],
                [296.55, 93.96, 314.97, 152.79],
                [328.94, 97.05, 342.49, 122.98],
                [356.62, 95.47, 372.33, 147.55],
                [464.08, 105.09, 495.74, 146.99],
                [276.11, 103.84, 291.44, 150.72],
            ],
            "scores": [0.546, 0.3, 0.407, 0.611, 0.335, 0.805, 0.953],
            "labels": ["4", "1", "0", "0", "0", "0", "0"],
        },
        {
            "boxes": [
                [72.92, 45.96, 91.23, 80.57],
                [45.17, 45.34, 66.28, 79.83],
                [82.28, 47.04, 99.66, 78.50],
                [59.96, 46.17, 80.35, 80.48],
                [75.29, 23.01, 91.85, 50.85],
                [71.14, 1.10, 96.96, 28.33],
                [61.34, 55.23, 77.14, 79.57],
                [41.17, 45.78, 60.99, 78.48],
                [56.18, 44.80, 64.42, 56.25],
            ],
            "scores": [
                0.532,
                0.204,
                0.782,
                0.202,
                0.883,
                0.271,
                0.561,
                0.204,
                0.349,
            ],
            "labels": ["49", "49", "49", "49", "49", "49", "49", "49", "49"],
        },
    ]

    return [
        Detection(
            uid=str(idx),
            groundtruths=[
                BoundingBox(
                    xmin=box[0],
                    ymin=box[1],
                    xmax=box[2],
                    ymax=box[3],
                    labels=[("class", label_value)],
                )
                for box, label_value in zip(gt["boxes"], gt["labels"])
            ],
            predictions=[
                BoundingBox(
                    xmin=box[0],
                    ymin=box[1],
                    xmax=box[2],
                    ymax=box[3],
                    labels=[("class", label_value)],
                    scores=[score],
                )
                for box, label_value, score in zip(
                    pd["boxes"], pd["labels"], pd["scores"]
                )
            ],
        )
        for idx, (gt, pd) in enumerate(zip(groundtruths, predictions))
    ]


@pytest.fixture
def false_negatives_single_datum_baseline_detections() -> list[Detection]:
    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                    scores=[0.8],
                ),
                BoundingBox(
                    xmin=100,
                    xmax=110,
                    ymin=100,
                    ymax=200,
                    labels=[("key", "value")],
                    scores=[0.7],
                ),
            ],
        )
    ]


@pytest.fixture
def false_negatives_single_datum_detections() -> list[Detection]:
    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                    scores=[0.8],
                ),
                BoundingBox(
                    xmin=100,
                    xmax=110,
                    ymin=100,
                    ymax=200,
                    labels=[("key", "value")],
                    scores=[0.9],
                ),
            ],
        )
    ]


@pytest.fixture
def false_negatives_two_datums_one_empty_low_confidence_of_fp_detections() -> list[
    Detection
]:

    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                    scores=[0.8],
                ),
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                    scores=[0.7],
                ),
            ],
        ),
    ]


@pytest.fixture
def false_negatives_two_datums_one_empty_high_confidence_of_fp_detections() -> list[
    Detection
]:

    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                    scores=[0.8],
                ),
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                    scores=[0.9],
                ),
            ],
        ),
    ]


@pytest.fixture
def false_negatives_two_datums_one_only_with_different_class_low_confidence_of_fp_detections() -> list[
    Detection
]:

    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                    scores=[0.8],
                ),
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "other value")],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                    scores=[0.7],
                ),
            ],
        ),
    ]


@pytest.fixture
def false_negatives_two_images_one_only_with_different_class_high_confidence_of_fp_detections() -> list[
    Detection
]:

    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                    scores=[0.8],
                ),
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "other value")],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=[("key", "value")],
                    scores=[0.9],
                ),
            ],
        ),
    ]



from valor_lite.detection import DataLoader, Detection


def test_metadata_using_torch_metrics_example(
    torchmetrics_detections: list[Detection],
):
    """
    cf with torch metrics/pycocotools results listed here:
    https://github.com/Lightning-AI/metrics/blob/107dbfd5fb158b7ae6d76281df44bd94c836bfce/tests/unittests/detection/test_map.py#L231
    """
    manager = DataLoader()
    manager.add_data(torchmetrics_detections)
    evaluator = manager.finalize()

    assert evaluator.ignored_prediction_labels == [("class", "3")]
    assert evaluator.missing_prediction_labels == []
    assert evaluator.n_datums == 4
    assert evaluator.n_labels == 6
    assert evaluator.n_groundtruths == 20
    assert evaluator.n_predictions == 19

    assert evaluator.metadata == {
        "ignored_prediction_labels": [
            ("class", "3"),
        ],
        "missing_prediction_labels": [],
        "n_datums": 4,
        "n_labels": 6,
        "n_groundtruths": 20,
        "n_predictions": 19,
    }
